#include <ctype.h>
#include <stddef.h>
#include "Global.h"
#include "Attack.h"
#include "Util.h"
#include "Object.h"
#include "Hero.h"
#include "FloppyDisk.h"
#include "OperatingSystem.h"

#ifdef _WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

static shObjectIlk
*randomFloppyDiskIlk (void)
{
    return AllIlks[RNG(kObjFirstFloppyDisk+1, kObjLastFloppyDisk)];
}

shCompInterface checkInterface (shObject *computer)
{
    switch (computer->mIlkId) {
    case kObjMiniComputer:    return kIVisualOnly;
    case kObjMegaComputer:    return kIVisualVoice;
    case kObjSatCom:          return kIVisualVoice;
    case kObjBioComputer:     return kIUniversal;
    default:                  return kNotAComputer;
    }
}

static void
confusedComputingMessage (void)
{
    if (Hero.isBlind ())
        if (Hero.isMute ())
            I->p ("You misexpress some of the commands in your confusion.");
        else
            I->p ("You mispronounce some of the commands in your confusion.");
    else
        I->p ("You make a few typos in your confusion.");
}

/* Checks whether hero can read map or object/monster locations printed
   on screen. Alternatively nerve interface can be employed in which
   case the image forms directly in the hero's brain.

returns:
   0: hero cannot get displayed information with specified interface
   1: hero managed to receive the information
*/
int visualDisplaySuccesful (shCompInterface interface) {
    if (!Hero.isBlind ()) {
        return 1;   /* Silent, this is normal situation. */
    } else if (interface != kIUniversal) {
        I->p ("But you can't see the display!");
        return 0;
    } else {
        I->p ("You receive information through your bio computer's hypodermic nerve interface.");
        return 1;
    }
}

void
criticalError (shObject *disk)
{
    I->p ("Critical error %d.  Execution halted.", RNG (1, 500));
    disk->setInfectedKnown ();
}

int
hackingRoll (shObject *computer)
{
    int roll = sportingD20 ();
    int sk = Hero.getSkillModifier (kHacking);
    /* Bonus from tools bundled with OS. */
    int os = systemHackingModifier (computer->mEnhancement);
    /* Drunk coding never good. */
    int conf = Hero.is (kConfused) ? -5 : 0;
    return roll + sk + os + conf;
}

shAttack *BuggyHypnosis =
    newBurstAttack (shAttack::kNoAttack, 0, 1, kMesmerizing, 12, 80);

shAttack *DebuggedHypnosis =
    newBurstAttack (shAttack::kNoAttack, 0, 1, kMesmerizing, 6, 40);

shAttack *OptimizedHypnosis =
    newBurstAttack (shAttack::kNoAttack, 0, 1, kMesmerizing, 5, 25);

shExecResult
doHypnosis (shObject *computer, shObject *disk)
{
    disk->setIlkKnown ();
    if (Hero.isBlind ()) {
        I->p ("You listen to some %s trance music.",
            disk->isInfected () ? "pitiful" : "excellent");
    } else {
        I->p ("%s displays a hypnotic screensaver!", YOUR (computer));
    }

    if (disk->isBuggy ()) {
        Level->areaEffect (BuggyHypnosis, NULL, Hero.mX, Hero.mY,
                           kNoDirection, &Hero, 0, 0);
    } else if (disk->isOptimized ()) {
        Level->areaEffect (OptimizedHypnosis, NULL, Hero.mX, Hero.mY,
                           kNoDirection, &Hero, 0, 0);
    } else {
        Level->areaEffect (DebuggedHypnosis, NULL, Hero.mX, Hero.mY,
                           kNoDirection, &Hero, 0, 0);
    }

    return kNormal;
}

extern shObjId consumeIlk (const char *);

/* Attempt to write a new program on a blank floppy disk. */
shExecResult
doBlank (shObject *computer, shObject *disk)
{
    char buf[100];
    char buf2[100];
    char *s;
    int dc;
    int score;
    shObjectIlk *newilk;

    disk->setIlkKnown ();
    I->p ("This disk is blank!");
    I->getStr (buf, 80, "What program do you want to write on it?");
    s = &buf[0];
    while (isspace (*s)) ++s;
    if (0 == strncasecmp ("floppy disk of ", s, 15)) {
        strncpy (buf2, s, 80);
    }
    snprintf (buf2, 100, "floppy disk of %s", s);
    buf2[99] = 0;
    newilk = AllIlks[consumeIlk (buf2)];

    if (Hero.is (kConfused)) {
        shObjectIlk *confilk;
        do {
            confilk = randomFloppyDiskIlk ();
        } while (confilk == disk->myIlk () || confilk == newilk);
        newilk = confilk;
        confusedComputingMessage ();
        if (RNG (10)) { /* usually successful */
            score = 100;
        } else {
            score = 0;
        }
    } else {
        if (!newilk) {
            I->p ("There is no such program!");
            return kNoExpiry;
        } else if (AllIlks[kObjBlankDisk] == newilk) {
            I->p ("Don't be ridiculous!");
            return kNoExpiry;
        }

        score = hackingRoll (computer);
        if (disk->isOptimized ()) {
            score += 2;
        } else if (disk->isBuggy ()) {
            score -= 10;
        }
        if (newilk->mFlags & kIdentified) {
            score += 4;
        } else if (SoftwareEngineer != Hero.mProfession) {
            score -= 4;
        }
        if (newilk == AllIlks[kObjMatterCompiler] && RNG (5)) {
            score -= 100; /* this is very hard to write */
        }
    }
    dc = 10 + newilk->mCost / 10;
    if (score >= dc) {
        disk = Hero.removeOneObjectFromInventory (disk);
        /* Measly developer excuse for liking colorful floppies. */
        /*if (disk->mIlk->mGlyph.mChar != newilk->mGlyph.mChar) {
            if (!Hero.isBlind ()) {
                I->p ("Disk is repainted according to ISO software coloring standard.");
            }
        }*/
        // Found it to be annoying long term.
        // Perhaps include this excuse in blank disk's description.
        disk->mIlkId = newilk->mId;
        if (0 == Hero.addObjectToInventory (disk)) {
            I->p ("The disk slips from your hands!");
            Level->putObject (disk, Hero.mX, Hero.mY);
        }
    } else {
        I->p ("Your hacking attempt was a failure.");
        return kDestruct;
    }
    if (newilk == AllIlks[kObjOperatingSystem]) {
        I->p ("This was a very exhausting task.");
        Hero.sufferDamage (OptimizedHypnosis);
    }
    return kNoExpiry;
}


void
identifyObjects (int howmany, int infected)
{
    shObject *obj;
    int i;
    int didsomething = 0;
    const char *prompt = "What do you want to identify?";

    while (howmany) {
        shMenu *idmenu = I->newMenu (prompt, shMenu::kCategorizeObjects);
        shObjectVector v;

        /* Infected disks reveal appearance. It is useful only for
           zen games but still something to keep in mind. */
        if (!infected) {
            unselectObjectsByFunction (&v, Hero.mInventory,
                                       &shObject::isIdentified);
        } else {
            unselectObjectsByFunction (&v, Hero.mInventory,
                                       &shObject::isAppearanceKnown);
        }
        if (0 == v.count ()) {
            I->p ("You have %s unidentified items.",
                  didsomething ? "no more" : "no");
            return;
        }
        if (didsomething) {
            I->pause ();
        }
        for (i = 0; i < v.count (); i++) {
            obj = v.get (i);
            idmenu->addPtrItem (obj->mLetter, obj->inv (), obj, obj->mCount);
        }
        idmenu->getPtrResult ((const void **) &obj);
        if (obj) {
            ++didsomething;
            --howmany;
            if (!infected) {
                obj->identify ();
            } else {
                obj->setAppearanceKnown ();
            }
            I->p ("%c - %s", obj->mLetter, obj->inv ());
            prompt = "What do you want to identify next?";
        }
        delete idmenu;
    }
}


shExecResult
doIdentify (shObject *computer, shObject *disk)
{
    int howmany = disk->isOptimized () ? RNG (1, 5) : 1;

    disk->setIlkKnown ();
    I->p ("This is an identify program.");
    I->pause ();
    identifyObjects (howmany, disk->isInfected ());
    if (howmany > 1)
        disk->setBugginessKnown ();
    if (howmany) Hero.reorganizeInventory ();
    return kNormal;
}


shExecResult
doSpam (shObject *computer, shObject *disk)
{
    static const char *spammsg [] = {
        "Warning!  Your warpspace connection is not optimized!",
        "Make thousands of buckazoids per cycle from the comfort of your own pod!",
        "Beautiful Arcturian babes are waiting to meet you!"
    };
    static const char *spamprompt [] = {
        "Transmit %d buckazoids to Meta-Net Systems for an upgrade?",
        "Upload $%d to Hyperpyramid Industries to buy business plan?",
        "Beam over your Arcturian bride for only %d buckazoids?"
    };

    int idx = RNG (3);
    int price = RNG (500) + 500;
    int sucker = 0;

    if (price > Hero.countMoney ()) {
        price = Hero.countMoney ();
    }
    I->p ("%s", spammsg[idx]);

    disk->setIlkKnown ();

    if (price < 50) {
        I->p ("Run this disk again when you've got some more buckazoids.");
        return kNoExpiry;
    }

    if (disk->isBuggy ()) {
        I->p ("Transmitting %d buckazoids through Zero-Click purchase plan...",
              price);
        disk->setBugginessKnown ();
        sucker = 1;
    } else { /* Give clue this disk is infected. */
        int newidx = idx;
        if (disk->isInfected ()) {
            /* Spam message and money receiver will not match. */
            while (newidx == idx)
                newidx = RNG (3);
        }
        if (I->yn (spamprompt[newidx], price)) {
            sucker = 1;
        } else {
            disk->setIlkKnown ();
        }
    }

    if (sucker) {
        Hero.loseMoney (price);
        I->p ("Thank you for your order.  "
              "Please wait 4-6 orbit cycles for delivery.");
    } else {
        I->p ("Perhaps you might be interested in some of our other products?");
    }
    return kNoExpiry;
}


shExecResult
doCorrupt (shObject *computer, shObject *disk)
{
    disk->setIlkKnown ();
    criticalError (disk);
    return kNoExpiry;
}

/* Detection programs have realtively minor punishment for having
   the disk infected. 1/3 chances for not reporting found item. */
shExecResult
doDetectObject (shObject *computer, shObject *disk)
{
    shCompInterface interface = checkInterface (computer);
    disk->setIlkKnown ();
    I->p ("You scan the area for objects.");
    int display = visualDisplaySuccesful (interface);
    int stack = 0;
    int objs = 0;
    int constr = 0;

    if (!display) I->p ("Getting objects list will have to suffice.");

    if (!Hero.is (kConfused)) {
        if (display) {
            for (int y = 0; y < MAPMAXROWS; ++y) {
                for (int x = 0; x < MAPMAXCOLUMNS; ++x) {
                    shObjectVector *v = Level->getObjects (x, y);
                    if (v) {
                        Level->forgetObject (x, y);
                    }
                    if (v && (!disk->isInfected () || RNG (3))) {
                        int besttype = kMaxObjectType;
                        shObject *bestobj = NULL;
                        for (int i = 0; i < v->count (); ++i) {
                            shObject *obj = v->get (i);
                            objs += obj->mCount;
                            if (disk->isOptimized ()) obj->setAppearanceKnown ();
                            if (obj->myIlk ()->mType < besttype) {
                                besttype = v->get (i) -> myIlk () -> mType;
                                bestobj = v->get (i);
                            }
                        }
                        ++stack;
                        Level->remember (x, y, bestobj->mIlkId);
                        Level->drawSq (x, y);
                    }
                    shCreature *c = Level->getCreature (x, y);
                    if (c && (!disk->isInfected () || RNG (3))
                          && c->myIlk ()->mType == kConstruct)
                    {
                        ++constr;
                        Level->remember (x, y, c->mIlkId);
                        Level->drawSq (x, y);
                    }
                }
            }
        }
        if (!disk->isBuggy ()) {
            for (int i = 0; i < Hero.mInventory->count (); ++i)
                if (!disk->isInfected () || RNG (3)) {
                    shObject *obj = Hero.mInventory->get (i);
                    obj->setAppearanceKnown ();
                }
        }
    } else if (display) {
        for (int i = 0; i < Level->mFeatures.count (); ++i) {
            shFeature *f = Level->mFeatures.get (i);
            if (!disk->isInfected () || RNG (3)) {
                if (f->isTrap () && !f->isBerserkDoor ())
                    continue;
                if ((f->mType == shFeature::kDoorHiddenVert ||
                    f->mType == shFeature::kDoorHiddenHoriz))
                {
                    if (disk->isBuggy ()) continue;
                    f->mType = shFeature::kDoorClosed;
                    f->mSportingChance = 0;
                    f->mTrapUnknown = 0;
                }
                if (f->isBerserkDoor () && disk->isOptimized ()) {
                    f->mTrapUnknown = 0;
                }
                Level->remember (f->mX, f->mY, f);
                Level->drawSq (f->mX, f->mY);
            }
        }
    }
    I->pauseXY (Hero.mX, Hero.mY);
    I->p ("%1$d object%2$s stack%2$s detected.",
        stack, (stack == 0 || stack > 1) ? "s" : "");
    I->p ("%d object%s detected.",
        objs, (objs == 0 || objs > 1) ? "s" : "");
    I->p ("%d construct%s detected.",
        constr, (constr == 0 || constr > 1) ? "s" : "");
    return kNormal;
}


/* Returns number of creatures detected. */
int
doDetectMonKind (shObject *computer, shObject *disk,
                int (shCreature::*predicate) ())
{
    shCompInterface interface = checkInterface (computer);
    int canGetInfo = visualDisplaySuccesful (interface);
    int cnt = 0;
    for (int y = 0; y < MAPMAXROWS; y++) {
        for (int x = 0; x < MAPMAXCOLUMNS; x++) {
            shCreature *c = Level->getCreature (x, y);
            if (c && (c->*predicate) () &&
                (!disk->isInfected () || RNG (3)))
            {
                if (canGetInfo) {
                    I->drawMem (x, y, kNone, c, NULL, NULL, NULL);
                }
                cnt++;
            }
        }
    }
    I->pauseXY (Hero.mX, Hero.mY);
    return cnt;
}


shExecResult
doDetectLife (shObject *computer, shObject *disk)
{
    disk->setIlkKnown ();
    if (Hero.is (kConfused)) {
        I->p ("You scan the area for droids.");
        if (disk->isBuggy ()) {
            I->pause ();
            I->p ("These aren't the droids you're looking for.");
            return kNormal;
        }
        I->pauseXY (Hero.mX, Hero.mY);
        int cnt = doDetectMonKind (computer, disk, &shCreature::isRobot);
        I->p ("%d droid%s detected.", cnt, (cnt == 0 || cnt > 1) ? "s" : "");
    } else {
        I->p ("You scan the area for lifeforms.");
        if (disk->isBuggy ()) { /* you detect yourself */
            I->p ("1 lifeform detected.");
            I->pauseXY (Hero.mX, Hero.mY);
            return kNormal;
        }
        I->pauseXY (Hero.mX, Hero.mY);
        int cnt = doDetectMonKind (computer, disk, &shCreature::isAlive);
        I->p ("%d lifeform%s detected.", cnt, (cnt == 0 || cnt > 1) ? "s" : "");
    }
    return kNormal;
}


shExecResult
doDetectTraps (shObject *computer, shObject *disk)
{
    int cnt = 0;
    int canGetInfo;
    shCompInterface interface = checkInterface (computer);

    I->p ("You scan the area for traps.");
    canGetInfo = visualDisplaySuccesful (interface);
    for (int i = 0; i < Level->mFeatures.count (); ++i) {
        shFeature *f = Level->mFeatures.get (i);
        if (f->isTrap () && (!disk->isInfected () || RNG (3))) {
            if (f->isBerserkDoor ()) {
                continue; /* not a trap, a malfunction */
            }
            if (canGetInfo) {
                f->mTrapUnknown = 0;
                Level->remember (f->mX, f->mY, f);
            }
            ++cnt;
        }
    }
    I->p ("%d trap%s detected.", cnt, (cnt == 0 || cnt > 1) ? "s" : "");
    return kNormal;
}


shExecResult
doDetectBugs (shObject *computer, shObject *disk)
{
    shObject *obj;
    int btotal = 0;
    int i;

    disk->setIlkKnown ();
    if (Hero.is (kConfused)) {
        I->pauseXY (Hero.mX, Hero.mY);
        I->p ("You scan the area for bugs.");
        btotal = doDetectMonKind (computer, disk, &shCreature::isInsect);
    } else if (disk->isBuggy ()) {
        disk->setBugginessKnown ();
        I->p ("This disk is buggy!");
        Hero.reorganizeInventory ();
        return kNormal;
    } else {
        I->p ("Scanning inventory for bugs...");
        for (i = 0; i < Hero.mInventory->count (); i++) {
            obj = Hero.mInventory->get (i);
            /* One third chance to ignore item. */
            if (!disk->isInfected () || RNG (3)) {
                if (obj->isBuggy ()) {
                    obj->setBugginessKnown ();
                    btotal++;
                } else if (disk->isOptimized ()) {
                    obj->setBugginessKnown ();
                }
            }
        }
    }
    I->p ("%d bugs found.", btotal);
    Hero.reorganizeInventory ();
    return kNormal;
}


shExecResult
doMapping (shObject *computer, shObject *disk)
{
    shCompInterface interface = checkInterface (computer);
    disk->setIlkKnown ();
    if (Hero.is (kConfused)) {
        return doDetectTraps (computer, disk);
    }
    I->p ("A map appears on your screen!");
    if (visualDisplaySuccesful (interface)) {
        if (!disk->isInfected ()) {
            Level->reveal (0);
        } else {
            Level->reveal (1);
            I->p ("It seems to be missing some parts.");
            disk->setDamageKnown ();
        }
    }
    if (computer->isA (kObjSatCom)) {
        int aliens = doDetectMonKind (computer, disk, &shCreature::isAlien);
        if (aliens) {
            I->p ("%d worthy prey found.", aliens);
        }
    }
    return kNormal;
}


shExecResult
doInfectedDebug (shObject *computer, shObject *disk)
{
    /* Pretend to debug debugged computer. */
    if (!computer->isBugginessKnown () && computer->isDebugged ()) {
        I->p ("%s seems to be working much better.", computer->your ());
        disk->setIlkKnown ();
        computer->setBugginessKnown ();
        return kNormal;
    }

    if (Hero.is (kConfused)) {
        /* Pretend to patch itself. */
        if (!disk->isBugginessKnown () && disk->isDebugged()) {
            if (disk->mCount > 1) {
                I->p ("You patch some bugs in your %d "
                "floppy disks of debugging.", disk->mCount);
            } else {
                I->p ("You patch some bugs in your floppy disk of debugging.");
            }
            disk->setIlkKnown ();
            disk->setBugginessKnown ();
            return kNormal;
        }

        /* Pretend to optimize itself. */
        if (!disk->isBugginessKnown () && disk->isOptimized ()) {
            if (disk->mCount > 1) {
                I->p ("You optimize your %d floppy disks of debugging.",
                    disk->mCount);
            } else {
                I->p ("You optimize your floppy disk of debugging.");
            }
            disk->setIlkKnown ();
            disk->setBugginessKnown ();
            return kNormal;
        }

        /* Pretend to debug some objects if hero is not aware of any
           buggy worn equipment and wears/wields some stuff with unknown
           bugginess status. */
        if ((disk->isBugginessKnown () && disk->isOptimized ())
            || !disk->isBugginessKnown ())
        {
            int i;
            int buggyAndKnown = 0;
            shObject *obj;
            shObjectVector v;
            selectObjectsByFunction (&v,
                Hero.mInventory, &shObject::isBugginessKnown);
            for (i = 0; i < v.count (); ++i) {
                obj = v.get (i);
                if (obj->isBuggy ()) {  /* Plan foiled. */
                    buggyAndKnown = 1;
                    break;
                }
            }
            if (!buggyAndKnown) {
                /* Good. Virus can not help and still pretend to be useful. */
                int unknownObjs = Hero.mInventory->count () - v.count ();
                if (unknownObjs) {
                    I->p ("%d objects debugged.", RNG (unknownObjs + 1));
                    disk->setIlkKnown ();
                    return kNormal;
                } else {
                    I->p ("0 objects debugged.");
                    return kNormal;
                }/* Sigh. Try something else. */
            }
        }   /* No way to cheat on optimized debugging. */
    } else if (!disk->isBugginessKnown () || !disk->isOptimized ()) {
        shObjectVector v1, v2, v3;
        selectObjectsByFunction (&v1, Hero.mInventory, &shObject::isWorn);
        if (Hero.mWeapon) v1.add (Hero.mWeapon);
        selectObjectsByFunction (&v2, &v1, &shObject::isDebugged);
        unselectObjectsByFunction (&v3, &v2, &shObject::isBugginessKnown);
        /* Pretend to debug debugged worn/wielded stuff. This works as long
           as there is at least one unknown debugged used thing. We do not
           necessarily need to worry about known used buggy items provided
           aforementioned condition holds. */
        if (v3.count ()) {
            shObject *obj = v3.get (RNG (v3.count ()));
            I->p ("Debugging %s", obj->an ());
            obj->setBugginessKnown ();
            disk->setIlkKnown ();
            return kNormal;
        } else { /* Now buggy known equipment needs not to exist. */
            shObjectVector v4, v5;
            selectObjectsByFunction (&v4, &v1, &shObject::isBuggy);
            selectObjectsByFunction (&v5, &v4, &shObject::isBugginessKnown);
            /* Pretend nothing to debug. */
            if (!v5.count ()) {
                disk->setIlkKnown ();
                I->p ("No showstopper bugs found.");
                return kNormal;
            }
        }
    }
    I->p ("No showstopper bugs found.");
    return kNormal;
}


int
canBeOptimized (shObject *obj)
{
    return !obj->isBugProof () &&
        (!obj->isOptimized () || !obj->isBugginessKnown ());
}


shExecResult
doDebug (shObject *computer, shObject *disk)
{   /* Infected debug is complex enough to get own subroutine. */
    if (disk->isInfected ())
        return doInfectedDebug (computer, disk);
    /* May need to reorganize inventory because objects with new state
       might be able to stack with others of its ilk. */
    int objsChanged = 0;
    shObject *obj;
    shObjectVector v;

    if (computer->isBuggy ()) {
        computer->setDebugged ();
        I->p ("%s seems to be working much better.", computer->your ());
        disk->setIlkKnown ();
        computer->setBugginessKnown ();
        return kNormal;
    }
    if (Hero.is (kConfused)) {
        disk->setIlkKnown ();
        disk->resetBugginessKnown ();
        if (disk->isBuggy ()) {
            disk->setDebugged ();
            I->p ("You patch some bugs in %s.", disk->your ());
            objsChanged = 1;
        } else if (disk->isOptimized ()) {
            int i;
            selectObjectsByFunction (&v, Hero.mInventory, &shObject::isBuggy);
            for (i = 0; i < v.count (); i++) {
                obj = v.get (i);
                obj->setDebugged ();
                if (obj->isA (kObjTorc) && obj->isWorn ()) Hero.torcCheck ();
            }
            I->p ("%d objects debugged.", i);
            objsChanged = i;
        } else {
            disk->setOptimized ();
            I->p ("You optimize %s.", disk->your ());
            objsChanged = 1;
        }
        disk->setBugginessKnown ();
        return kNormal;
    } else if (disk->isOptimized ()) {
        shObjectVector v;
        selectObjectsByCallback (&v, Hero.mInventory, canBeOptimized);
        if (0 == v.count ()) {
            I->p ("You have nothing to debug or optimize.");
            return kNormal;
        }
        disk->setIlkKnown ();
        disk->setBugginessKnown ();
        I->p ("You've found a floppy disk of debugging!");
        obj = Hero.quickPickItem (&v, "debug or optimize", shMenu::kCategorizeObjects);
        if (NULL == obj) {
            return kNormal;
        }
        if (obj->isBugProof ()) {
            I->p ("%s %s not appear to be affected.", obj->your (),
                obj->mCount > 1 ? "do" : "does");
            return kNormal;
        } else if (obj->isBuggy ()) {
            obj->setDebugged ();
            I->p ("You patch some bugs in %s.", obj->your ());
            objsChanged = 1;
        } else if (obj->isOptimized ()) {
            I->p ("You verify that %s is fully optimized.", obj->your ());
        } else {
            obj->setOptimized ();
            I->p ("You optimize %s.", obj->your ());
            objsChanged = 1;
        }
        if (obj->isA (kObjTorc) && obj->isWorn ()) Hero.torcCheck ();
        obj->setBugginessKnown ();
        if (objsChanged) Hero.reorganizeInventory();
        return kNormal;
    } else {
        shObjectVector v2;
        selectObjectsByFunction (&v, Hero.mInventory, &shObject::isWorn);
        selectObjectsByFunction (&v2, &v, &shObject::isBuggy);
        if (Hero.mWeapon && Hero.mWeapon->isBuggy ()) {
            v2.add (Hero.mWeapon);
        }
        if (v2.count ()) {
            obj = v2.get (RNG (v2.count ()));
            obj->setDebugged ();
            obj->resetBugginessKnown ();
            I->p ("Debugging %s.", obj->an ());
            obj->setBugginessKnown ();
            if (obj->isA (kObjTorc) && obj->isWorn ()) Hero.torcCheck ();
            disk->setIlkKnown ();
            objsChanged = 1;
        } else {
            disk->setIlkKnown ();
            I->p ("No showstopper bugs found.");
        }
        if (objsChanged) {
            Hero.reorganizeInventory ();
            Hero.computeIntrinsics ();
            Hero.computeSkills ();
        }
        return kNormal;
    }
}


shExecResult
doEnhanceArmor (shObject *computer, shObject *disk)
{
    shObject *obj;
    shObjectVector v, w;

    selectObjects (&v, Hero.mInventory, kArmor);
    selectObjectsByFunction (&w, &v, &shObject::isWorn);
    v.reset ();
    if (!Hero.is (kConfused)) {
        selectObjectsByFunction (&v, &w, &shObject::isEnhanceable);
    } else {
        selectObjectsByFunction (&v, &w, &shObject::isFooproofable);
    }
    if (0 == v.count ()) {
        I->p ("Your skin tingles for a moment.");
        disk->setIlkKnown ();
        return kNormal;
    }
    obj = v.get (RNG (v.count ()));
    shObjectIlk *ilk = obj->myIlk ();
    const char *your_armor = obj->your ();
    int max_opt_enhc = ilk->mMaxEnhancement;
    int max_dbg_enhc = ilk->mMaxEnhancement / 2 + ilk->mMaxEnhancement % 2;

    if (Hero.is (kConfused)) {
        if (kNoEnergy != obj->vulnerability ()) {
            if (!obj->isFooproof ()) {
                if (!disk->isInfected ()) {
                    obj->setFooproof ();
                    obj->setFooproofKnown ();
                } else {
                    criticalError (disk);
                    return kNormal;
                }
            } else {
                obj->setFooproofKnown ();
            }
        }
        I->p ("%s vibrates.", your_armor);
        disk->setIlkKnown ();
        if (obj->mDamage) {
            if (!disk->isInfected ()) {
                obj->mDamage = 0;
                if (!Hero.isBlind ()) {
                    I->p ("%s looks as good as new!", your_armor);
                }
            } else {
                criticalError (disk);
                return kNormal;
            }
        }
    } else if (disk->isBuggy ()) {
        if (obj->mEnhancement > -max_opt_enhc) {
            --obj->mEnhancement;
        }
        obj->setBuggy ();
        disk->setBugginessKnown ();
        if (!Hero.isBlind ()) {
            I->p ("Blue smoke billows from %s.", your_armor);
            obj->setBugginessKnown ();
        } else {
            I->p ("You smell smoke.");
        }
    } else if (disk->isInfected ()) {
        int stealth = 0;
        /* Make sure increasing enhancement can be avoided. */
        if (!disk->isBugginessKnown () || disk->isDebugged ()) {
            stealth = (obj->mEnhancement >= max_dbg_enhc);
        } else if (disk->isOptimized ()) {
            stealth = obj->mEnhancement == max_opt_enhc;
        }
        /* Make sure debugging/optimizing armor can be avoided. */
        if (stealth) {
            if (!obj->isBugginessKnown () || obj->isOptimized ()) {
                stealth = 1;
            } else if (obj->isDebugged ()) {
                stealth = !disk->isBugginessKnown () || disk->isDebugged ();
            } else {
                stealth = 0;
            }
        }
        if (stealth) {
            I->p ("%s feels warm for an instant.", your_armor);
            disk->setIlkKnown ();
        } else {
            criticalError (disk);
            return kNormal;
        }
    } else if (disk->isOptimized ()) {
        int bonus = (max_opt_enhc + 1 - obj->mEnhancement) / 2;
        bonus = bonus <= 0 ? 0 : RNG (1, bonus);
        obj->mEnhancement += bonus;
        if (!obj->isOptimized () && obj->isBugginessKnown ())
            disk->setBugginessKnown ();
        obj->setOptimized ();
        if (bonus >= 1) {
            disk->setIlkKnown ();
            disk->setBugginessKnown ();
            obj->setBugginessKnown ();
        }
        I->p ("%s feels warm for %s.", your_armor, bonus > 1 ? "a while" :
              bonus > 0 ? "a moment" : "an instant");
    } else {
        disk->setIlkKnown ();
        if (obj->mEnhancement < max_dbg_enhc) {
            ++obj->mEnhancement;
            I->p ("%s feels warm for a moment.", your_armor);
        } else {
            I->p ("%s feels warm for an instant.", your_armor);
            disk->setBugginessKnown ();
        }
        if (obj->isBuggy ()) {
            obj->setDebugged ();
        }
    }
    Hero.computeAC ();
    return kNormal;
}


shExecResult
doEnhanceImplant (shObject *computer, shObject *disk)
{
    shObject *obj;
    shObjectVector v, w;

    disk->setIlkKnown ();
    selectObjects (&v, Hero.mInventory, kImplant);
    selectObjectsByFunction (&w, &v, &shObject::isWorn);
    v.reset ();
    selectObjectsByFunction (&v, &w, &shObject::isEnhanceable);
    if (0 == v.count ()) {
        I->p ("Your brain itches for a moment.");
        return kNormal;
    }
    obj = v.get (RNG (v.count ()));

    if (disk->isBuggy ()) {
        I->p ("You feel a stinging sensation in your %s.",
              describeImplantSite (obj->mImplantSite));
        --obj->mEnhancement;
        obj->setBuggy ();
        obj->setBugginessKnown ();
        disk->setBugginessKnown ();
    } else if (disk->isInfected ()) {
        int stealth = 0;
        /* Make sure increasing enhancement can be avoided. */
        if (!obj->isEnhancementKnown ()) {
            stealth = 1;
        } else if (!disk->isBugginessKnown () || disk->isDebugged ()) {
            stealth = obj->mEnhancement >= 3;
        } else if (disk->isOptimized ()) {
            stealth = obj->mEnhancement == 5;
        }
        if (stealth) {
            I->p ("You feel a brief tingling sensation in your %s.",
              describeImplantSite (obj->mImplantSite));
        } else {
            criticalError (disk);
        }
    } else if (disk->isOptimized ()) {
        int bonus = (6 - obj->mEnhancement) / 2;
        bonus = bonus <= 0 ? 0 : RNG (1, bonus);
        obj->mEnhancement += bonus;
        I->p ("You feel a %s sensation in your %s.",
              bonus > 1 ? "buzzing" :
              bonus > 0 ? "tingling" : "brief tingling",
              describeImplantSite (obj->mImplantSite));
    } else if (obj->mEnhancement < 3) {
        ++obj->mEnhancement;
        I->p ("You feel a tingling sensation in your %s.",
              describeImplantSite (obj->mImplantSite));
    } else {
        I->p ("You feel a brief tingling sensation in your %s.",
              describeImplantSite (obj->mImplantSite));
    }
    disk->setIlkKnown ();
    Hero.computeIntrinsics ();
    return kNormal;
}


shExecResult
doEnhanceWeapon (shObject *computer, shObject *disk)
{
    shObject *obj;
    shObjectVector v;
    const char *your_weapon;

    disk->setIlkKnown ();

    if (NULL == Hero.mWeapon) {
        I->p ("Your hands sweat profusely.");
        return kNormal;
    }

    obj = Hero.mWeapon;
    your_weapon = obj->your ();
    shObjectIlk *ilk = obj->myIlk ();
    int hasattacks = ilk->mMeleeAttack || ilk->mMissileAttack ||
                     ilk->mGunAttack || ilk->mZapAttack;
    if (!hasattacks || obj->isBugProof ()) {
        I->p ("Your hands sweat profusely.");
        return kNormal;
    }
    const char *third_person = obj->mCount > 1 ? "" : "s";
    int isPhaser = obj->isA (kObjPhaser);
    if (Hero.is (kConfused)) {
        if (kNoEnergy != obj->vulnerability ()) {
            if (!obj->isFooproof ()) {
                if (!disk->isInfected ()) {
                    obj->setFooproof ();
                    obj->setFooproofKnown ();
                } else {
                    criticalError (disk);
                    return kNormal;
                }
            } else {
                obj->setFooproofKnown ();
            }
        }
        I->p ("%s vibrate%s.", your_weapon, third_person);
        if (obj->mDamage) {
            if (!disk->isInfected ()) {
                obj->mDamage = 0;
                if (!Hero.isBlind ()) {
                    I->p ("%s look%s as good as new!", your_weapon, third_person);
                }
            } else {
                criticalError (disk);
                return kNormal;
            }
        }
    } else if (disk->isBuggy ()) {
        I->p ("%s shudder%s%s!", your_weapon, third_person,
              obj->isBuggy () ||
              !(obj->isMeleeWeapon () && obj->isAimedWeapon ()) ?
                  "" : " and welds to your hand");
        if (obj->isEnhanceable ()) --obj->mEnhancement;
        obj->setBuggy ();
        obj->setBugginessKnown ();
        disk->setBugginessKnown ();
    } else if (disk->isInfected ()) {
        int stealth = 0;
        /* Make sure increasing enhancement can be avoided. */
        if (obj->isEnhancementKnown ()) {
            if (!disk->isBugginessKnown () || disk->isDebugged ()) {
                if (isPhaser) {
                    stealth = obj->mEnhancement >= 4;
                } else {
                    stealth = obj->mEnhancement >= 3;
                }
            } else if (disk->isOptimized ()) {
                if (isPhaser) {
                    stealth = obj->mEnhancement == 6;
                } else {
                    stealth = obj->mEnhancement == 5;
                }
            }
        } else {
            stealth = 1;
        }
        /* Make sure debugging/optimizing weapon can be avoided. */
        if (stealth) {
            if (!obj->isBugginessKnown () || obj->isOptimized ()) {
                stealth = 1;
            } else if (obj->isDebugged ()) {
                stealth = !disk->isBugginessKnown () || disk->isDebugged ();
            } else {
                stealth = 0;
            }
        }
        if (stealth) {
            if (isPhaser && disk->isOptimized () && disk->isBugginessKnown ()) {
                I->p ("You fine tune the phase variance for %s.", your_weapon);
            } else {
                I->p ("%s feel%s warm for an instant.", your_weapon, third_person);
            }
        } else {
            criticalError (disk);
        }
    } else if (disk->isOptimized ()) {
        if (!obj->isOptimized () && obj->isBugginessKnown ())
            disk->setBugginessKnown ();
        obj->setOptimized ();
        if (!obj->isEnhanceable ()) {
            I->p ("%s feel%s warm for an instant.", your_weapon, third_person);
        } else {
            int bonus;
            if (isPhaser)
                bonus = (7 - obj->mEnhancement) / 2;
            else
                bonus = (6 - obj->mEnhancement) / 2;
            bonus = bonus <= 0 ? 0 : RNG (1, bonus);
            if (bonus > 1) disk->setEnhancementKnown ();
            obj->mEnhancement += bonus;
            if (isPhaser) {
                I->p ("You %s the phase variance for %s.",
                bonus > 1 ? "modify" :
                bonus > 0 ? "tune" : "fine tune", your_weapon);
            } else {
                I->p ("%s feel%s warm for %s.", your_weapon, third_person,
                    bonus > 1 ? "a while" :
                    bonus > 0 ? "a moment" : "an instant");
            }
        }
    } else {
        if (!obj->isEnhanceable ()) {
            I->p ("%s feel%s warm for an instant.", your_weapon, third_person);
        } else if (obj->mEnhancement < 3) {
            ++obj->mEnhancement;
            if (isPhaser) {
                I->p ("You modify %s phase variance.", your_weapon);
            } else {
                I->p ("%s feel%s warm for a moment.", your_weapon, third_person);
            }
        } else {
            if (isPhaser && (obj->mEnhancement == 3)) {
                ++obj->mEnhancement;
                I->p ("You finish fine-tuning %s phase variance.", your_weapon);
            } else {
                I->p ("%s feel%s warm for an instant.", your_weapon, third_person);
            }
        }
        if (obj->isBuggy ()) {
            obj->setDebugged ();
        }
    }
    return kNormal;
}


shWarpTrace traces[MAXTRACES];
int lastTrace;


shExecResult
doWarpTrace (shObject *computer, shObject *disk)
{
    if (Level->noTransport ()) {
        I->p ("Nothing happens.");
        disk->maybeName ();
        return kNormal;
    }

    if (disk->isBuggy ()) {
        if (Level->attractWarpMonsters (Hero.mX, Hero.mY)) {
            disk->setIlkKnown ();
            disk->setBugginessKnown ();
        } else {
            I->p ("Nothing happens.");
            disk->maybeName ();
        }
        return kNormal;
    }
    disk->setIlkKnown ();
    int found = 0;
    for (int i = 0; i < MAXTRACES; ++i) {
        /* Give traces from oldest to freshest. */
        int idx = (lastTrace + i) % MAXTRACES;
        if (traces[idx].level == Hero.mLevel &&
            traces[idx].time + 20000 >= Clock)
        {
            int dist = distance (&Hero, traces[idx].destX, traces[idx].destY);
            I->p ("%s transported to a %s location.", traces[idx].who,
                dist < 60 ? "neighboring" :
                dist < 120 ? "nearby" : "distant");
            ++found;
        }
    }
    if (!found) {
        I->p ("No warp traces found.");
    }
    disk->setBugginessKnown ();
    if (found && disk->isOptimized ()) {
        int last = (lastTrace + MAXTRACES - 1) % MAXTRACES;
        if (I->yn ("Follow last trace?")) {
            int x = -1, y = -1;
            if (Hero.is (kConfused)) {
                x = traces[last].srcX;
                y = traces[last].srcY;
            } else {
                x = traces[last].destX;
                y = traces[last].destY;
            }
            Level->findNearbyUnoccupiedSquare (&x, &y);
            Hero.transport (x, y, 100, 1);
        }
    }
    return kNormal;
}

/* returns 0 on success
           1 if the creature dies
           -1 if otherwise failed
*/
int
shCreature::transport (int x, int y, int tries, int safe, int override)
{
    int res;
    int oldX = mX, oldY = mY;

    if (-1 == x) {
        mLevel->findUnoccupiedSquare (&x, &y);
    }
    if (isHero ()) {
        if (Level->noTransport () && !override) {
            I->p ("You %s for a moment.",
                Hero.isBlind () ? "shudder" : "flicker");
            return 0;
        }
    }
    mTrapped = 0;
    mDrowning = 0;
    while (tries--) {
        res = mLevel->moveCreature (this, x, y);
        if (-1 == res) {
            /* Try again, random location. */
            mLevel->findUnoccupiedSquare (&x, &y);
        } else {
            /* Register this transport. */
            traces[lastTrace].srcX = oldX;
            traces[lastTrace].srcY = oldY;
            traces[lastTrace].destX = mX;
            traces[lastTrace].destY = mY;
            traces[lastTrace].level = mLevel;
            if (traces[lastTrace].who) free (traces[lastTrace].who);
            traces[lastTrace].who = strdup (myIlk ()->mName);
            traces[lastTrace].time = Clock;
            lastTrace = (lastTrace + 1) % MAXTRACES;
            /* Monsters with kWarpish feat might follow. */
            if (isHero ())
                Level->attractWarpMonsters (x, y);
            return res;
        }
    }
    if (!safe) {
        die (kTransporterAccident);
        return 1;
    }
    return -1;
}


shExecResult
doTransport (shObject *computer, shObject *disk)
{
    int x = -1, y = -1;

    disk->setIlkKnown ();
    if (disk->isBuggy () && !RNG (20)) {
        /* Failed transport. */
        I->p ("The process fails.");
        disk->setBugginessKnown ();
        return kNormal;
    } else if (disk->isOptimized () && !Hero.is (kConfused)) {
        /* controlled local transport */
        if (I->getSquare ("Transport to what location?", &x, &y, -1)) {
            if (Level->isMainframe () || disk->isInfected ()) {
                int tries = 50;
                int nx, ny;
                do {
                    nx = x + RNG (15) - 8;
                    ny = y + RNG (9) - 4;
                    Level->findNearbyUnoccupiedSquare (&nx, &ny);
                } while (nx == x && ny == y && tries--);
                I->p ("It's hard to transport accurately in here.");
                x = nx; y = ny;
            }
            Hero.transport (x, y, 100, 1);
        }
        return kNormal;
    } else {
        int tries = disk->isInfected () ? 10 : 100;
        /* random local transport */
        Level->findSquare (&x, &y);
        Hero.transport (x, y, tries, 1);
        return kNormal;
    }
}

/* Gives only part of a string. */
const char *
shHero::radLevels (void)
{
    return
        Hero.mRad > 1200 ? "unbelieveable" :
        Hero.mRad >  500 ? "extreme" :
        Hero.mRad >  300 ? "lethal" :
        Hero.mRad >  200 ? "very dangerous" :
        Hero.mRad >  125 ? "dangerous" :
        Hero.mRad >   75 ? "significant" :
        Hero.mRad >   50 ? "moderate" :
        Hero.mRad >   25 ? "acceptable" :
                           "minimal";
}

/* List hidden properties of a creature. */
void
shCreature::doDiagnostics (int quality, int present, /* 1 */ int file /* -1 */)
{
    struct shGrammar
    {
        char you[4]; char r[5]; char are[9]; char have[9];
    };
    const shGrammar declension[5] =
    {
        {    "You",     "Your",   "You are",  "You have" },
        {    "He",      "His",    "He is",    "He has"   },
        {    "She",     "Her",    "She is",   "She has"  },
        {    "It",      "Its",    "It is",    "It has"   },
        {    "You",     "Your",   "You were", "You had"  }
    };
#   define YOU offsetof (shGrammar, you)
#   define YOU_R offsetof (shGrammar, r)
#   define YOU_ARE offsetof (shGrammar, are)
#   define YOU_HAVE offsetof (shGrammar, have)
    static const struct shDiagTestData
    {
        const char *fmt;
        unsigned int offset;
        int (shCreature::*predicate) ();
    } tests[] =
    {
        { "%s lucky.",                      YOU_ARE,    &shCreature::isLucky },
        { "%s very acute scent sense.",     YOU_HAVE,   &shCreature::hasScent },
        { "%s can swim.",                   YOU,        &shCreature::canSwim },
        { "%s terrifying presence.",        YOU_HAVE,   &shCreature::isScary },
        { "%s telepathic.",                 YOU_ARE,    &shCreature::hasTelepathy },
        { "%s tremor sense.",               YOU_HAVE,   &shCreature::hasTremorsense },
        { "%s EM field vision.",            YOU_HAVE,   &shCreature::hasEMFieldVision },
        { "%s night vision.",               YOU_HAVE,   &shCreature::hasNightVision },
        { "%s X-ray vision.",               YOU_HAVE,   &shCreature::hasXRayVision },
        { "%s reflection.",                 YOU_HAVE,   &shCreature::hasReflection },
        { "%s narcoleptic.",                YOU_ARE,    &shCreature::hasNarcolepsy },
        { "%s a light source.",             YOU_ARE,    &shCreature::hasLightSource },
        { "%s shielded.",                   YOU_ARE,    &shCreature::hasShield },
        { "%s psionic protection.",         YOU_HAVE,   &shCreature::hasBrainShield },
        { "%s an air supply.",              YOU_HAVE,   &shCreature::hasAirSupply },
        { "%s automatic searching.",        YOU_HAVE,   &shCreature::hasAutoSearching },
        { "%s breathless.",                 YOU_ARE,    &shCreature::isBreathless },
        { "%s sessile.",                    YOU_ARE,    &shCreature::isSessile },
        { "%s able to multiply rapidly.",   YOU_ARE,    &shCreature::isMultiplier },
        { "%s attracted to warp traces.",   YOU_ARE,    &shCreature::isWarpish },
        { "%s body eliminates radiation.",  YOU_R,      &shCreature::hasRadiationProcessing },
        { "%s body regenerates quickly.",   YOU_R,      &shCreature::hasAutoRegeneration },
        { "%s highly acidic blood.",        YOU_HAVE,   &shCreature::hasAcidBlood },
        { "%s monitor your health.",        YOU,        &shCreature::hasHealthMonitoring },
        { "%s will explode if destroyed.",  YOU,        &shCreature::isExplosive },
        { "%s able to understand alien languages.", YOU_ARE, &shCreature::hasTranslation }
    };
    static const struct shDiagVarData
    {
        char fmt[50];
        unsigned int offset;
        int shCreature::*member;
    } vars[] =
    {
        { "%s speed %s %d.",                YOU_R, &shCreature::mSpeed },
        { "%s speed bonus %s %d.",          YOU_R, &shCreature::mSpeedBoost },
        { "%s natural armor bonus %s %d.",  YOU_R, &shCreature::mNaturalArmorBonus },
        { "%s to hit modifier %s %d." ,     YOU_R, &shCreature::mToHitModifier },
        { "%s damage modifier %s %d." ,     YOU_R, &shCreature::mDamageModifier },
        { "%s psionic modifier %s %d." ,    YOU_R, &shCreature::mPsiModifier }
    }; /* How many first variables reveal even if value is zero. */
#   define ALWAYS_SHOW 1
    const int BUFSIZE = 70;
    char buf[BUFSIZE];
    const shGrammar *you;
    if (!present) {
        you = &declension[4];
    } else if (isHero ()) {
        you = &declension[0];
    } else if (hasMind () || mGender == kMale) {
        you = &declension[1];
    } else if (hasMind () || mGender == kFemale) {
        you = &declension[2];
    } else {
        you = &declension[3];
    }
    shMenu *menu = I->newMenu ("Personal diagnostics", shMenu::kNoPick);
    /* Report variables first. */
    for (unsigned i = 0; i < sizeof (vars) / sizeof (shDiagVarData); ++i) {
        if (i < ALWAYS_SHOW || this->*vars[i].member) {
            snprintf (buf, BUFSIZE, vars[i].fmt,
                (char *) ((char *) you + vars[i].offset),
                (present) ? "is" : "was", this->*vars[i].member);
            if (file == -1) {
                menu->addText (buf);
            } else {
                write (file, buf, strlen (buf));
                write (file, "\n", 1);
            }
        }
    } /* Resistance or vulnerability to energy types. */
    int e;
    int resist;
    const char *description;
    for (e = kNoEnergy; e < kMaxEnergyType; e++) {
        resist = getResistance ((shEnergyType) e);
        description = energyDescription ((shEnergyType) e);
        if (resist && description) {
            snprintf (buf, BUFSIZE, "%s %s to %s.", you->are,
                resist < 0  ? "particularly vulnerable" :
                resist < 5  ? "somewhat resistant" :
                resist < 10 ? "resistant" :
                resist < 50 ? "extremely resistant" :
                              "nearly invulnerable",
                description);
            if (file == -1) {
                menu->addText (buf);
            } else {
                write (file, buf, strlen (buf));
                write (file, "\n", 1);
            }
        }
    } /* Test for intrinsics. */
    for (unsigned i = 0; i < sizeof (tests) / sizeof (shDiagTestData); ++i) {
        if (tests[i].predicate == &shCreature::hasXRayVision &&
            this->mInnateIntrinsics & kBlind)
        {   /* Naturally blind characters do not benefit from X-ray vision. */
            continue;
        }
        if ((this->*tests[i].predicate) ()) {   /* Present. */
            /* Default report line. */
            snprintf (buf, BUFSIZE, tests[i].fmt,
                (char *) ((char *) you + tests[i].offset));
            if (!present) { /* Past tense clumsiness. */
                /* This is not pretty! Got no better idea now. :( -- MB */
                char *mod;
                if (strstr (buf, "swim")) {
                    snprintf (buf, BUFSIZE, "You could swim.");
                } else if ((mod = strstr (buf, "tes"))) {
                    *(mod + 2) = 'd'; /* eliminated, regenerated */
                } else if (strstr (buf, "monitor")) {
                    snprintf (buf, BUFSIZE,
                        "You monitored your health.");
                } else if (strstr (buf, "explode")) {
                    snprintf (buf, BUFSIZE, "You exploded.");
                }
            }
            if (file == -1) {
                menu->addText (buf);
            } else {
                write (file, buf, strlen (buf));
                write (file, "\n", 1);
            }
        }
    } /* Vaguely express rad levels. */
    if (isHero ()) { /* Not implemented for monsters. */
        snprintf (buf, BUFSIZE, "%s been exposed to %s levels of radiation.",
                  you->have, Hero.radLevels ());
        if (file == -1) {
            menu->addText (buf);
        } else {
            write (file, buf, strlen (buf));
            write (file, "\n", 1);
        }
    } /* Only hero may have this --> only second person form needed. */
    if (isHero ()) {
        int tmp;
        if ((tmp = Hero.getStoryFlag ("impregnation"))) {
            if (tmp > 25 || !present) {
                snprintf (buf, BUFSIZE, "The alien lifeform inside you"
                    " %s %skill you%s.", present ? "will" : "would",
                    tmp > 25 ? "soon " : "", present ? "" : " anyway");
            } else {
                snprintf (buf, BUFSIZE, "You've been impregnated with "
                    "a dangerous alien lifeform.");
            }
            if (file == -1) {
                menu->addText (buf);
            } else {
                write (file, buf, strlen (buf));
                write (file, "\n", 1);
            }
        }
        if (Hero.getStoryFlag ("superglued tongue")) {
            snprintf (buf, BUFSIZE, "The canister of super glue %s eventually"
                " come off.", (present) ? "will" : "would");
            if (file == -1) {
                menu->addText (buf);
            } else {
                write (file, buf, strlen (buf));
                write (file, "\n", 1);
            }
        }
    }
    if (file == -1) menu->finish ();
    delete menu;
}


shExecResult
doDiagnostics (shObject *computer, shObject *disk)
{
    disk->setIlkKnown ();
    I->p ("Initiating body scan.");
    I->pauseXY (Hero.mX, Hero.mY);
    /* Bugginess and infectedness each decrease quality by one point. */
    Hero.doDiagnostics (disk->mBugginess - disk->isInfected ());
    return kNormal;
}

static int
isHackable (shObject *obj)
{
    return obj->isA (kFloppyDisk) || obj->isA (kObjGenericKeycard);
}

static int
isBioHackable (shObject *obj)
{
    return isHackable (obj) ||
        (obj->isIlkKnown () && obj->isA (kObjRestorationCanister));
}

shExecResult
doHacking (shObject *computer, shObject *disk)
{
    shObject *obj;
    shObjectVector v;

    /* Do not blow computer virus' cover by revealing true nature
       of this disk in upcoming choice of floppy to hack. */
    if (disk->mIlkId != kObjComputerVirus) {
        disk->setIlkKnown ();
    }
    if (!disk->isIlkKnown () || !disk->isAppearanceKnown ()) {
        I->p ("You've found a floppy disk of hacking!");
    }
    if (computer->isA (kObjBioComputer)) {
        selectObjectsByCallback (&v, Hero.mInventory, isBioHackable);
    } else {
        selectObjectsByCallback (&v, Hero.mInventory, isHackable);
    }
    obj = Hero.quickPickItem (&v, "hack", 0);
    if (!obj) return kNormal;

    if (obj->isA (kObjBlankDisk)) {
        I->p ("No program there.");
        obj->setIlkKnown ();
        return kNormal;
    }
    if (obj->isCracked () && !obj->isCrackedKnown ()) {
        obj->setCrackedKnown ();
        if (obj->isA (kFloppyDisk)) {
            I->p ("This software has already been cracked!");
        } else if (obj->isA (kObjGenericKeycard)) {
            I->p ("The access level of this keycard has already been hacked!");
        } else {
            I->p ("This thingy has already been cracked!");
        }
        return kNormal;
    }
    int docrack = 1;
    if (obj->isInfected () && obj->isInfectedKnown ()) {
        docrack = !I->yn ("Do you wish to remove virus manually?");
    }

    int score = hackingRoll (computer);
    if (obj->isA (kFloppyDisk)) {
        if (disk->isOptimized ()) {
            score += 2;
        } else if (disk->isBuggy ()) {
            score -= 10;
        }
        if (obj->mIlkId == kObjHackingDisk || obj->mIlkId == kObjMatterCompiler) {
            score -= 100; /* this is nearly impossible */
        }
        if (obj->isOptimized () || obj->isBuggy ()) {
            score -= 2; /* debugged code is more straightforward to hack */
        }
        if (!obj->isIlkKnown ()) {
            score -= 2; /* you have no clue what it is supposed to do */
        }
        if (obj->isInfected ()) {
            score -= 2; /* yet another complication */
        }
        if (obj->isInfected () && !obj->isInfectedKnown ()) {
            score -= 2; /* and you still have to learn about it */
        }
        if (docrack) {
            if (score >= 15 && !disk->isInfected ()) {
                if (!obj->isCracked ()) {
                    I->p ("You remove the copy protection from %s.", obj->your ());
                    obj->setCracked ();
                    obj->setCrackedKnown ();
                } else {
                    I->p ("You reenable the copy protection of %s.", obj->your ());
                    obj->resetCracked ();
                    obj->setCrackedKnown ();
                }
                if (score >= 19) {
                    if (obj->isInfected () && !obj->isInfectedKnown ()) {
                        I->p ("You also notice trojan code interspersed among program routines.");
                    }
                    obj->setInfectedKnown ();
                }
                return kNormal;
            }
        } else {
            if (score >= 20 && !disk->isInfected ()) {
                I->p ("You remove the virus from %s.", obj->your ());
                obj->resetInfected ();
                obj->setCrackedKnown ();
                if (score >= 24) {
                    if (obj->isCracked () && !obj->isCrackedKnown ()) {
                        I->p ("You also notice this software is cracked.");
                    }
                    obj->setCrackedKnown ();
                }
                return kNormal;
            }
        }
        I->p ("Your hacking attempt was a failure!");
        if (docrack) obj->setCrackedKnown ();
        if (score <= 10) {
            obj->setBuggy ();
        }
    } else if (obj->isA (kObjGenericKeycard)) {
        I->p ("You attach %s to USB port of %s.", THE (obj), YOUR (computer));
        if (obj->isCracked () && !obj->isCrackedKnown ()) {
            obj->setCrackedKnown ();
            I->p ("This keycard has already been cracked!");
            return kNormal;
        }
        if (score >= 15 && !disk->isInfected ()) {
            if (!obj->isCracked ()) {
                I->p ("You set highest level of access on %s.", YOUR (obj));
                obj->identify (); /* Identify, since you learned its codes. */
                obj->setCracked ();
            } else {
                I->p ("You reset access level to original state on %s.", YOUR (obj));
                obj->resetCracked ();
            }
            if (obj->isA (kObjMasterKeycard)) {
                I->p ("It is a master keycard so you did it for sport.");
            }
        } else {
            I->p ("You failed to find access level data for %s.", YOUR (obj));
            if (obj->isA (kObjMasterKeycard)) {
                obj->setIlkKnown ();
                I->p ("It is a master keycard. No point in this anyway.");
            }
        }
    } else if (obj->isA (kObjRestorationCanister)) {
        I->p ("You submerge %s in %s.", THE (obj), YOUR (computer));
        obj = Hero.removeOneObjectFromInventory (obj);
        if (score >= 25 && !disk->isInfected ()) {
            I->p ("You remove limitations on nanomachines in %s.", YOUR (obj));
            obj->mIlkId = kObjGainAbilityCanister;
        } else if (score <= 15 || disk->isInfected ()) {
            I->p ("You fail to remove limitations on nanomachines correctly.");
            obj->mIlkId = kObjPoisonCanister;
        } else {
            I->p ("You fail to remove limitations on nanomachines.");
        }
        Hero.addObjectToInventory (obj);
    } else {
        I->p ("Error POPCORN DELTA!");
        I->p ("You somehow chosen object for hacking without associated routine.");
        I->p ("Please report this bug.");
    }
    return kNormal;
}

void
doInstallRawSystem (shObject *computer, shObject *disk, int speak)
{
    if (disk->isOptimized ()) {
        computer->mEnhancement = RNG(kAverageSys, kSuperbSys);
    } else if (disk->isBuggy ()) {
        computer->mEnhancement = RNG(kAbysmalSys, kAverageSys);
    } else {
        computer->mEnhancement = RNG(kPoorSys, kGoodSys);
    }
    if (Hero.mProfession != SoftwareEngineer) {
        computer->resetEnhancementKnown ();
    } else if (!Hero.isBlind ()) {
        /* Programmer must have watched installation to know this. */
        computer->setEnhancementKnown ();
    }
    if (speak) {
        I->p ("Operating system installed successfully.");
    }
}

/* Generic stuff. */
#define SYS_UPGRADE "Upgrading system architecture."
#define SYS_DOWNGRADE "Downgrading system architecture."
/* Do we need better excuse for capping this? -- MB */
#define SYS_DEBUGGED_CAP \
    I->p ("This lite system distribution does not contain all packages."); \
    I->p ("Please upgrade to deluxe system distribution.")
/* Omitting semicolon above so I can have one after the macro. */
shExecResult
doInstallSystem (shObject *computer, shObject *disk)
{
    disk->setIlkKnown ();
    if (Hero.is (kConfused)) {
        I->p ("You scan the area for programs.");
        if (disk->isBuggy ()) {
            I->p ("Remote repository does not include your favorite programs.");
            I->pauseXY (Hero.mX, Hero.mY);
            return kNormal;
        }
        I->pauseXY (Hero.mX, Hero.mY);
        int cnt = doDetectMonKind (computer, disk, &shCreature::isProgram);
        I->p ("%d program%sdetected.", cnt, (cnt == 0 || cnt > 1) ? "s " : " ");
        return kNormal;
    }

    shObject *obj;
    shObjectVector v;
    selectObjectsByFunction (&v, Hero.mInventory, &shObject::isComputer);
    obj = Hero.quickPickItem (&v, "install or upgrade system", 0);
    if (obj == NULL)
        return kNormal;
    int hadSystem = (obj->mEnhancement != kNoSys);
    /* Rule for this is: if specific module was installed the system
       quality is known precisely. */
    int revealSysQuality = 0;

    if (disk->isBuggy ())   /* Buggy disks attempt downgrade. */
        disk->setBugginessKnown ();

    if (disk->isInfected ()) {
        if (!computer->isEnhancementKnown ()) {
            if (disk->isBuggy ()) {
                I->p (SYS_DOWNGRADE);
            } else {
                I->p (SYS_UPGRADE);
            }
        } else {
            criticalError (disk);
        }
        return kNormal;
    }
    switch (obj->mEnhancement) {
        case kNoSys:
            doInstallRawSystem (obj, disk, 1);
            break;
        case kAbysmalSys:
            if (disk->isBuggy ()) {
                I->p ("Further system downgrade not possible.");
                disk->setBugginessKnown ();
                revealSysQuality = 1;
            } else {
                I->p ("Installing voice support module.");
                ++obj->mEnhancement;
                /* Does not reveal bugginess because player has no idea
                   whether this disk would get him past kGoodSys. */
                revealSysQuality = 1;
            }
            break;
        case kTerribleSys:
            if (disk->isBuggy ()) {
                I->p ("Uninstalling voice support module.");
                --obj->mEnhancement;
                disk->setBugginessKnown ();
                revealSysQuality = 1;
            } else {
                I->p (SYS_UPGRADE);
                ++obj->mEnhancement;
            }
            break;
        case kBadSys:
            if (disk->isBuggy ()) {
                I->p (SYS_DOWNGRADE);
                --obj->mEnhancement;
                disk->setBugginessKnown ();
            } else {
                I->p (SYS_UPGRADE);
                ++obj->mEnhancement;
            }
            break;
        case kPoorSys:
            if (disk->isBuggy ()) {
                I->p (SYS_DOWNGRADE);
                --obj->mEnhancement;
                disk->setBugginessKnown ();
            } else {
                I->p ("Installing hypodermic nerve interface support.");
                ++obj->mEnhancement;
                revealSysQuality = 1;
            }
            break;
        case kAverageSys:
            if (disk->isBuggy ()) {
                I->p ("Uninstalling hypodermic nerve interface support.");
                --obj->mEnhancement;
                disk->setBugginessKnown ();
                revealSysQuality = 1;
            } else {
                I->p ("Installing transparent update architecture.");
                ++obj->mEnhancement;
                revealSysQuality = 1;
            }
            break;
        case kFineSys:
            if (disk->isBuggy ()) {
                I->p ("Uninstalling transparent update architecture.");
                --obj->mEnhancement;
                disk->setBugginessKnown ();
                revealSysQuality = 1;
            } else {
                I->p ("Installing system hardware analyzer.");
                ++obj->mEnhancement;
                revealSysQuality = 1;
            }
            break;
        case kGoodSys:
            if (disk->isBuggy ()) {
                I->p ("Uninstalling system hardware analyzer.");
                --obj->mEnhancement;
                revealSysQuality = 1;
            } else if (disk->isOptimized ()){
                I->p ("Installing software reliability tester.");
                disk->setBugginessKnown ();
                ++obj->mEnhancement;
                revealSysQuality = 1;
            } else {
                SYS_DEBUGGED_CAP;
                disk->setBugginessKnown ();
            }
            break;
        case kExcellentSys:
            if (disk->isBuggy ()) {
                I->p ("Uninstalling software reliability tester.");
                --obj->mEnhancement;
                revealSysQuality = 1;
            } else if (disk->isOptimized ()){
                I->p (SYS_UPGRADE);
                ++obj->mEnhancement;
            } else {
                SYS_DEBUGGED_CAP;
                disk->setBugginessKnown ();
            }
            break;
        case kSuperbSys:
            if (disk->isBuggy ()) {
                I->p (SYS_DOWNGRADE);
                --obj->mEnhancement;
            } else {
                I->p ("Further system upgrade not possible.");
                revealSysQuality = 1;
            }
            break;
    }

    if (hadSystem) {
        if (revealSysQuality) {
            obj->setEnhancementKnown ();
        }
    }
    return kNormal;
}


shExecResult
doAntiVirus (shObject *computer, shObject *disk)
{
    if (computer->isInfected ()) {
        computer->resetInfected ();
        computer->setDamageKnown ();
        I->p ("%s seems to be working much better.", computer->your ());
        disk->setIlkKnown ();
        return kNormal;
    }

    disk->setIlkKnown ();
    if (Hero.is (kConfused)) {
        shObjectVector v;
        shObject *obj;
        selectObjectsByFunction (&v, Hero.mInventory, &shObject::isInfected);
        if (!disk->isBuggy ()) {
            for (int i = 0; i < v.count (); ++i) {
                obj = v.get (i);
                if (!disk->isBuggy ()) {
                    obj->setDamageKnown ();
                }
                if (disk->isOptimized ()) {
                    obj->mDamage = 0;
                    if (disk->mIlkId == kObjComputerVirus) {
                        obj->mIlkId = kObjBlankDisk;
                    }
                }
            }
        }
        I->p ("Found %s%d infected items.",
            disk->isOptimized () ? "and recovered " : "", v.count ());
        if (disk->isOptimized ()) {
            Hero.reorganizeInventory ();
        }
    } else if (!computer->isFooproof ()) {
        I->p ("Installing antivirus.");
        computer->setFooproof ();
        computer->setFooproofKnown ();
        computer->setDamageKnown ();
    } else {
        I->p ("You verify that %s is protected.", computer->your ());
    }
    return kNormal;
}

#define MAXDRIVES 6
static int
countFloppyDrives (shObject *computer)
{
    switch (computer->mIlkId) {
    case kObjMiniComputer: return 1;
    case kObjMegaComputer: return MAXDRIVES;
    case kObjBioComputer: return 3;
    case kObjSatCom: return 2;
    default: return 0;
    }
}

/* Returns blank disk. */
shObject *
doSingleFloppyReformat (shObject *disk)
{
    shObject *newdisk = createObject ("blank floppy disk", 0);
    newdisk->mBugginess = disk->mBugginess;
    if (disk->isBugginessKnown ()) {
        newdisk->setBugginessKnown ();
    }
    newdisk->setInfectedKnown ();
    Hero.useUpOneObjectFromInventory (disk);
    return newdisk;
}

/* Found idea for this from Cyrus' scratchpad and there it was called
   floppy disk of reformat. I went with it instead of more staightforward
   floppy disk of formatting. -- MB */
shExecResult
doReformat (shObject *computer, shObject *disk)
{
    int drives = countFloppyDrives (computer);
    int i, chosen;
    char drivechar;
    shObject *inDrive[MAXDRIVES];
    shObject *obj;
    shObjectVector v;

    I->p ("This is a reformat program.");
    disk->setIlkKnown ();
    disk->setBugginessKnown ();

    inDrive[0] = disk;
    for (i = 1; i < drives; ++i)
        inDrive[i] = NULL;

    if (disk->isBuggy ()) { /* Acts immediately. */
        chosen = RNG (drives + 1);
        drivechar = 'A' + chosen;
        debug.log ("buggy reformat phase");
        /* Place disk at C for single drive computers. */
        if (chosen == drives && drives == 1)
            ++drivechar;
        /* Picked empty floppy drive. */
        if (chosen != drives && inDrive[chosen] == NULL) {
            int retry = 1;
            do { /* TODO: Use dropResult to avoid recreating the menu. */
                int choice;
                shMenu *DOSJoke = I->newMenu ("Abort, retry or fail?", 0);
                DOSJoke->addIntItem ('a', "abort", 'a', 1);
                DOSJoke->addIntItem ('r', "retry", 'r', 1);
                DOSJoke->addIntItem ('f', "fail",  'f', 1);
                /* First message below should not end with a dot.
                   MS-DOS did not have one there. -- MB */
                I->p ("Not ready reading drive %c", drivechar);
                I->p ("Abort, Retry, Fail?");
                I->pause ();
                /* Forced choice. Not picking anything will ask again. */
                DOSJoke->getIntResult (&choice);
                delete DOSJoke;
                if (choice == 'a') {
                    return kDestruct; /* Abort meant lost work. */
                } else if (choice == 'f') {
                    retry = 0;
                }
            } while (retry);
        } else {
            I->p ("Reformat complete.");
            if (RNG (2)) {
                doSingleFloppyReformat (disk);
                debug.log ("disk picked itself");
                /* No blow up since disk is now blank. No eat either since
                   original disk pointer is now invalid. */
                return kGoneAlready;
            } else {
                debug.log ("disk picked computer's HDD");
                computer->mEnhancement = kNoSys;
                return kNormal;
            }
        }
    }

    debug.log ("nonbuggy reformat phase");
    if (disk->isInfected ()) { /* Free hint. */
        I->p ("Warning: failed CRC data check.");
        disk->setDamageKnown ();
    }
    int count, choices;
    char prompt[50];
    if (disk->isOptimized ()) {
        choices = drives - 1;
    } else {
        choices = mini(1, drives - 1);
    }
    snprintf (prompt, 50, "You may choose %d floppy disk%s for reformatting:",
        choices, (choices > 1) ? "s" : "");

    shMenu *disks = I->newMenu (prompt,
        (choices > 1) ? shMenu::kMultiPick | shMenu::kCountAllowed : 0);
    if (choices != 0) { /* No need to build the menu if not going to ask. */
        debug.log ("collect reformatable floppies phase");
        int added = 0;
        selectObjects (&v, Hero.mInventory, kFloppyDisk);
        for (i = 0; i < v.count (); ++i) {
            obj = v.get (i);
            /* Original reformat disk is already in, so ... */
            if (obj == disk) {
                if (obj->mCount != 1) {
                    disks->addPtrItem (obj->mLetter, obj->inv (), obj,
                        obj->mCount - 1);
                    ++added;
                }
            } else {
                disks->addPtrItem (obj->mLetter, obj->inv (), obj, obj->mCount);
                ++added;
            }
        }
        if (!added) {   /* Hero only posesses reformat disk. */
            choices = 0;
        }
    }

    snprintf (prompt, 50, "Select drive%s to be reformatted:",
        (choices > 1) ? "s" : "" );
    shMenu *drivesToFormat = I->newMenu (prompt,
        disk->isOptimized() ? shMenu::kMultiPick : 0);

    i = 0;
    drivechar = 'a';
    debug.log ("assign floppy drives phase");
    I->pause ();
    while (choices != 0 && disks->getPtrResult ((const void**) &obj, &count)) {
        if (count == 0) {
            return kNormal;
        }
        /* For multiple reformat all drives with disks except the
         * first are preselected. For single reformat no preselection.
         * Note that while blind reformat disk may not be in first (unselected)
         * drive and get formatted if player accepts without looking.
         */
        while (count && i < drives) {
            if (inDrive[i] != NULL) {   /* At most one to skip. */
                debug.log ("drive[%d]: %p", i, disk);
                ++i;
            }
            --count;
            if (choices == 1) { /* No point in filling remaining drives. */
                count = 0;
            }
            debug.log ("drive[%d]: %p", i, obj);
            inDrive[i++] = obj;
        }
    }
    delete disks;
    if (choices == 0) {   /* Stick reformat disk in the only floppy drive. */
        debug.log ("drive[%d]: %p", i, disk);
        inDrive[i++] = disk;
    }

    /* Blind people usually stick floppies in drives at random. */
    if ((Hero.isBlind () && RNG(3)) || Hero.is (kConfused)) {
        debug.log ("shuffle floppies in drives phase");
        shuffle (inDrive, drives, sizeof(shObject *));
    }

    debug.log ("build floppy drive menu phase");
    for (i = 0; i < drives; ++i) {
        if (inDrive[i]) {
            obj = inDrive[i];
            drivesToFormat->addPtrItem (drivechar++,
                obj->getDescription (), obj, 1, i > 0 && choices != 1);
        } else {
            drivesToFormat->addPtrItem (drivechar++, "empty", NULL, 1, 0);
        }
    } /* HDD letter is always one letter futher apart from floppy drives. */
    drivesToFormat->addPtrItem (++drivechar, "hard disk", NULL, -1, 0);

    debug.log ("reformat floppies phase");
    int reformatDiskReformatted = 0;
    int blankDisksCount = 0;
    shObject *blankDisks[MAXDRIVES];
    while (drivesToFormat->getPtrResult ((const void**) &obj, &count)) {
        if (count == -1) {
            computer->mEnhancement = kNoSys;
        } else if (obj != NULL) {
            if (obj == disk) {
                reformatDiskReformatted = 1;
                debug.log ("reformat reformats itself");
            }
            debug.log ("formatting %p", obj);
            blankDisks[blankDisksCount++] = doSingleFloppyReformat (obj);
        }
    }
    delete drivesToFormat;

    I->p ("Reformat complete.");
    /* Since reformat routine operates on multiple floppy disks it must
       handle disk eating by buggy computers itself. Only disks acutally
       reformatted and original disks may be devoured. */
    if (computer->isBuggy ()) {
        int eatHowMany = RNG (-blankDisksCount, +blankDisksCount);
        debug.log ("devour floppies phase: eat %d of %d",
            eatHowMany, blankDisksCount);
        for (i = 0; i < eatHowMany; ++i) {
            if (blankDisks[i] == disk) {
                /* Just a way to say no eating original disk since it
                   is already gone. FIXME: pick better variable name */
                reformatDiskReformatted = 1;
            }   /* Crunch. Crunch. Crunch. */
            debug.log ("devouring %p", blankDisks[i]);
            Hero.useUpOneObjectFromInventory (blankDisks[i]);
        }
        if (eatHowMany > 0) {    /* Oh no! */
            if (!Hero.isBlind ()) {
                I->p ("Your computer ate %d of inserted floppy disks!",
                    eatHowMany);
            } else {
                I->p ("You listen to a %s om-nom-nom sound%c",
                    /* This compurer could not eat more? */
                    ((eatHowMany == drives) ?
                    /* No computer could eat more?! */
                        ((drives == MAXDRIVES)  ?
                            "terrible" : "scary") : "suspicious"),
                    ((eatHowMany == drives) ? '!' : '.'));
            }
        } /* Give back the rest. Watch out not to start at negative index. */
        for (i = maxi(0, eatHowMany); i < blankDisksCount; ++i) {
            Hero.addObjectToInventory (blankDisks[i], 0);
        }
    } else {
        for (i = 0; i < blankDisksCount; ++i) {
            Hero.addObjectToInventory (blankDisks[i], 0);
        }
    }
    Hero.reorganizeInventory ();
    debug.log ("reformat returns");
    return reformatDiskReformatted ? kGoneAlready : kNoEat;
}


shExecResult
doComputerVirus (shObject *computer, shObject *disk)
{
    disk->setInfected (); /* Paranoia. This disk already should be infected. */
    if (computer->isFooproof ()) {
        /* No message. Antivirus already has printed one. */
        return kDestruct; /* You could have reformatted it. Too late. */
    } else {
        computer->setInfected ();
        if (computer->isFooproofKnown ()) {
            I->p ("You notice unusual system unstability.");
        }
        disk->setIlkKnown ();
        return kFakeExpiry;
    }
}

/* To check whether a Smart-Disc can fly back. */
extern int calcShortestPath (int, int, int, int, shMonId, shDirection [], int);

static void
putAwayDiscs (shObjectVector *v)
{
    for (int i = 0; i < v->count (); ++i) {
        shObject *obj = v->get (i);
        Level->putObject (obj, obj->mX, obj->mY);
    }
}

shExecResult
doRecallSmartDisc (shObject *computer, shObject *disk)
{
    disk->setIlkKnown ();
    if (disk->isBuggy ()) {
        I->p ("Software error.  Recall failed.");
        disk->setBugginessKnown ();
        return kNoExpiry;
    }
    /* Find all non-buggy Smart-Discs on floor.  Buggy discs do not respond. */
    const int MAXPATH = 30;
    shObjectVector discs;
    for (int x = 0; x < MAPMAXCOLUMNS; ++x) {
        for (int y = 0; y < MAPMAXROWS; ++y) {
            shObjectVector *v = Level->getObjects (x, y);
            if (v && v->count ()) {
                for (int i = v->count () - 1; i >= 0; --i) {
                    shObject *obj = v->get (i);
                    if (obj->isA (kObjSmartDisc) && !obj->isBuggy ()) {
                        v->remove (obj);
                        discs.add (obj);
                    }
                }
            }
        }
    }
    int n = discs.count ();
    if (n == 0) {
        I->p ("No Smart-Discs present.");
        return kNoExpiry;
    }
    shDirection buf[MAXPATH];
    /* Check distances for discs. */
    int *dist = new int[n];
    int reachable = 0;
    for (int i = 0; i < n; ++i) {
        shObject *disc = discs.get (i);
        /* Uses kMonProgram because it means no flying past closed doors. */
        dist[i] = calcShortestPath (disc->mX, disc->mY, Hero.mX, Hero.mY,
                                    kMonProgram, buf, MAXPATH);
        if (dist[i])  ++reachable;
    }
    if (!reachable) {
        if (n > 1 && disk->isOptimized ()) {
            I->p ("All of %d Smart-Discs are unreachable.", n);
        } else {
            I->p ("No reachable Smart-Discs present.");
        }
        delete [] dist;
        putAwayDiscs (&discs);
        return kNoExpiry;
    }
    /* Find disc with shortest path to fly. */
    int closest = 0;
    for (int i = 1; i < n; ++i) { /* Dist[x] == 0 means no path exists. */
        if (dist[i] && dist[i] < dist[closest])
            closest = i;
    }
    shObject *disc = discs.get (closest);
    /* Get the disc. */
    if (Hero.addObjectToInventory (disc, 1)) {
        if (!Hero.mWeapon) {
            Hero.wield (disc, 1);
        }
        I->p ("%c - %s", disc->mLetter, disc->inv ());
    } else {
        if (!Hero.isBlind ()) {
            I->p ("%s lands at your feet.", disc->an ());
        } else {
            I->p ("You hear a thud near your feet.");
        }
        Level->putObject (disc, Hero.mX, Hero.mY);
    }
    if (n > 1 && disk->isOptimized ()) {
        I->p ("There are %d more Smart-Disc%s.  %d of those are reachable.",
              n-1, n-1 > 1 ? "s" : "", reachable-1);
    }
    discs.remove (disc);
    putAwayDiscs (&discs);
    delete [] dist;
    return kNoExpiry;
}


void
attractLawyers ()
{
    int i;
    int x, y;
    int cnt = 0;

    for (i = 0; i < Level->mCrList.count () ; i++) {
        shCreature *c = Level->mCrList.get (i);
        if (c->isA (kMonLawyer)) {
            if (c->canSee (&Hero)) {
                ((shMonster *) c) -> makeAngry ();
            } else if (cnt < 3) {
                x = Hero.mX;
                y = Hero.mY;
                if (0 == Level->findAdjacentUnoccupiedSquare (&x, &y)) {
                    c->transport (x, y, 1, 1);
                    c->cure (kFleeing);
                    cnt++;
                }
            }
        }
    }

    if (!cnt && !RNG (13)) {
        x = Hero.mX;
        y = Hero.mY;
        if (0 == Level->findAdjacentUnoccupiedSquare (&x, &y)) {
            shMonster *m = new shMonster (kMonLawyer);
            if (0 == Level->putCreature (m, x, y)) {
                cnt++;
            }
        }
    }

    I->p ("You seem to have attracted the attention of %s lawyer%s.",
          cnt > 1 ? "some" : "a", cnt > 1 ? "s" : "");
}


/* Big enough to get own file. MatterCompiler.cpp */
extern shExecResult doMatterCompiler (shObject *computer, shObject *disk);

/* returns number of ms taken */
int
executeFloppyDisk (shObject *computer, shObject *disk)
{
    shObjectIlk *ilk = disk->myIlk ();
    shCompInterface interface = checkInterface (computer);
    int sys = computer->mEnhancement;
    shExecResult result;
    int summonlawyer = 0;
    int mishandle = 0;  /* User confused and system got fooled. */

    if (!hasSystem (computer)) {
        if (disk->mIlkId == kObjOperatingSystem) {
            disk->setIlkKnown ();
            /* Blind heroes will not know whether system installation
               really succeeded until they try the computer again. */
            if (Hero.isBlind ()) {
                I->p ("The seeking sound continues. "
                    "This disk must be bootable.");
            }
            if (!disk->isOptimized ()) {
                if (!Hero.isBlind ()) {
                    I->p ("Unsupported executable format.");
                }
                return HALFTURN;
            } else {
                if (!Hero.isBlind ()) {
                    I->p ("Booting from floppy.....success!");
                    disk->setBugginessKnown ();
                }
                doInstallRawSystem(computer, disk, !Hero.isBlind ());
                if (disk->isInfected ()) {
                    computer->setInfected (); /* gratis */
                }
                return computer->isOptimized () ? FULLTURN : LONGTURN;
            }
        }
        return HALFTURN;   /* not a system disk */
    }

    if (Hero.isBlind ()) {
        /* blind and mute */
        if (Hero.isMute ()) {
            if (interface == kIUniversal) {
                I->p ("You let in your bio computer's hypodermic nerve interface.");
                I->p ("You feel dirty.");
                if (!systemSupportsInterface (sys, kIUniversal)) {
                    I->p ("To add insult to injury this interface is not supported.");
                    return FULLTURN;
                }
            } else {
                I->p ("But you're blind and mute!");
                return 0;
            }
        /* blindness only */
        } else if (interface >= kIVisualVoice) {
            I->p ("You make use of your %s's voice interface.",
                computer->myIlk ()->mName);
            if (!systemSupportsInterface (sys, kIVisualVoice)) {
                I->p ("Or rather you try to because it is unsupported.");
                return FULLTURN;
            }
        } else {
            I->p ("But you're blind, and this computer doesn't "
                  "have a voice interface.");
            return 0;
        }
    }

    /* Inspired by Vista OS. */
    if (isUpdateTime (sys)) {
        systemUpdateMessage ();
        return FULLTURN;
    }

    if (!systemCheckDisk (computer, disk))
        return FULLTURN;

    /* Invoke secondary disk effect unless system mishandles this. */
    if (Hero.is (kConfused) && disk->mIlkId != kObjBlankDisk) {
        confusedComputingMessage ();
        /* abysmal: 50% terrible: 33.(3)% bad: 25% */
        if (sys < kPoorSys && !RNG (1, sys + 5)) {
            I->p ("The computer fails to correct your errors.");
            mishandle = 1;
        }
    }

    /* If we are here hero managed to run the stuff. */
    if (!mishandle) disk->setAppearanceKnown ();

    if (Hero.isInShop ()) {
        Hero.usedUpItem (disk, 1, "use");
        disk->resetUnpaid ();
    }

    /* Save infectedness because inserted disk may be gone after execution. */
    int disk_infected = disk->isInfected ();

    if (!mishandle) {
        result = ilk->mRunFunc (computer, disk);
    } else {
        result = kNormal;
    }

    if (disk_infected && !computer->isFooproof ()) {
        computer->setInfected ();
    } else if (result != kGoneAlready && !disk->isInfected () &&
        computer->isInfected () && disk->mIlkId != kObjAntivirus)
    {
        disk->setInfected ();
    }

    if (result & kDestroyComputer) {
        int ms = computer->isOptimized () ? FULLTURN : LONGTURN;
        Hero.useUpOneObjectFromInventory (disk);
        Hero.useUpOneObjectFromInventory (computer);
        return ms;
    }

    if (!(result & kGoneAlready)) {
        /* Check for police awareness of using cracked stuff. */
        if (!(result & kNoExpiry) && disk->isCracked ()) {
            int danger = Hero.getStoryFlag ("software piracy");
            if (0 == danger) {
                Hero.setStoryFlag ("software piracy", 5);
            } else if (RNG (0, danger) && RNG (0, danger)) {
                summonlawyer = 1;
            }
            if (!disk->isCrackedKnown ()) {
                I->p ("The licence for this software has been cracked!");
                disk->setCrackedKnown ();
            }
        }

        if (result == kDestruct) {
            I->p ("The floppy disk self destructs!");
            Hero.useUpOneObjectFromInventory (disk);
            disk = NULL;
        } else if (!(result & kNoExpiry) &&
                   !disk->isCracked () &&
                   (!RNG (3) || result & kFakeExpiry))
        {
            I->p ("Your license for this software has expired.");
            I->p ("The floppy disk self destructs!");
            Hero.useUpOneObjectFromInventory (disk);
            disk = NULL;
        } else if (computer->isBuggy () && !(result & kNoEat)) {
            if (RNG (2)) {
                if (Hero.isBlind ())
                    I->p ("You hear your computer go nyam-nyam.");
                else
                    I->p ("Your computer eats the disk!");
                Hero.useUpOneObjectFromInventory (disk);
                disk = NULL;
                computer->setBugginessKnown ();
            } else if (RNG (2)) {
                disk->setBuggy ();
            }
        }
    } else {
        disk = NULL;
    }

    if (summonlawyer) attractLawyers ();

    return computer->isOptimized () ? FULLTURN : LONGTURN;
}

int
shObject::runFloppyDisk (shObject *computer)
{   /* Caller ought to check for Hero.is (kUnableCompute). */
    Hero.employedItem (computer, 40 BZ);
    return executeFloppyDisk (computer, this);
}

void
initializeFloppyDisks ()
{
    /* Warp traces also get initialized here. */
    for (int i = 0; i < MAXTRACES; ++i) {
        traces[i].level = NULL;
        traces[i].who = NULL;
    }
    lastTrace = 0;

    int n = 0;

    /* each price group should have one doozy */

    /* enum: kObjGenericFloppyDisk */
    /* enum: kObjFirstFloppyDisk = kObjGenericFloppyDisk */
    NEW_FLOPPY_DISK(kObjGenericFloppyDisk,
        "floppy disk", "floppy disk", kGray, NULL, 0, ABSTRACT);
    AllIlks[kObjGenericFloppyDisk]->mParent = kObjNothing;

    /* enum: kObjBugDetectionDisk */
    NEW_FLOPPY_DISK(kObjBugDetectionDisk,
        "floppy disk of bug detection",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doDetectBugs, 25, 50);
    ++n;
    /* enum: kObjIdentifyDisk */
    NEW_FLOPPY_DISK(kObjIdentifyDisk,
        "floppy disk of identify",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doIdentify, 25, 150);
    ++n;
    /* enum: kObjSpamDisk */
    NEW_FLOPPY_DISK(kObjSpamDisk,
        "floppy disk of spam",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doSpam, 25, 150);
    /* enum: kObjCorruptDisk */
    NEW_FLOPPY_DISK(kObjCorruptDisk,
        "corrupt floppy disk", "unlabeled floppy disk",
        kGray,
        &doCorrupt, 25, 5);



    /* enum: kObjBlankDisk */
    NEW_FLOPPY_DISK(kObjBlankDisk,
        "blank floppy disk", "unlabeled floppy disk",
        kGray,
        &doBlank, 50, 35);
    ++n;
    /* enum: kObjObjectDetectionDisk */
    NEW_FLOPPY_DISK(kObjObjectDetectionDisk,
        "floppy disk of object detection",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doDetectObject, 50, 50);
    ++n;
    /* enum: kObjLifeformDetectionDisk */
    NEW_FLOPPY_DISK(kObjLifeformDetectionDisk,
        "floppy disk of lifeform detection",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doDetectLife, 50, 50);
    ++n;
    /* enum: kObjMappingDisk */
    NEW_FLOPPY_DISK(kObjMappingDisk,
        "floppy disk of mapping",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doMapping, 50, 70);
    ++n;
    /* enum: kObjDiagnosticsDisk */
    NEW_FLOPPY_DISK(kObjDiagnosticsDisk,
        "floppy disk of diagnostics",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doDiagnostics, 50, 20);
    ++n;
    /* enum: kObjWarpTracerDisk */
    NEW_FLOPPY_DISK(kObjWarpTracerDisk,
        "floppy disk of warp tracer",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doWarpTrace, 50, 30);


    ++n;
    /* enum: kObjDebuggerDisk */
    NEW_FLOPPY_DISK(kObjDebuggerDisk,
        "floppy disk of debugging",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doDebug, 100, 100);
    ++n;
    /* enum: kObjEnhanceArmorDisk */
    NEW_FLOPPY_DISK(kObjEnhanceArmorDisk,
        "floppy disk of enhance armor",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doEnhanceArmor, 100, 60);
    ++n;
    /* enum: kObjEnhanceWeaponDisk */
    NEW_FLOPPY_DISK(kObjEnhanceWeaponDisk,
        "floppy disk of enhance weapon",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doEnhanceWeapon, 100, 70);
    ++n;
    /* enum: kObjEnhanceImplantDisk */
    NEW_FLOPPY_DISK(kObjEnhanceImplantDisk,
        "floppy disk of enhance implant",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doEnhanceImplant, 100, 50);
    ++n;
    /* enum: kObjHypnosisDisk */
    NEW_FLOPPY_DISK(kObjHypnosisDisk,
        "floppy disk of hypnosis",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doHypnosis, 100, 50);
    ++n;
    /* enum: kObjTransportDisk */
    NEW_FLOPPY_DISK(kObjTransportDisk,
        "floppy disk of transportation",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doTransport, 100, 75);


    ++n;
    /* enum: kObjReformatDisk */
    NEW_FLOPPY_DISK(kObjReformatDisk,
        "floppy disk of reformat",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doReformat, 200, 60);
    ++n;
    /* enum: kObjComputerVirus */
    NEW_FLOPPY_DISK(kObjComputerVirus,
        "floppy disk of computer virus",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doComputerVirus, 200, 50);
    ++n;
    /* enum: kObjRecallDisk */
    NEW_FLOPPY_DISK(kObjRecallDisk,
        "floppy disk of Smart-Disc recall",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doRecallSmartDisc, 200, 20);


    ++n;
    /* enum: kObjOperatingSystem */
    NEW_FLOPPY_DISK(kObjOperatingSystem,
        "floppy disk of operating system",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doInstallSystem, 300, 30);
    ++n;
    /* enum: kObjHackingDisk */
    NEW_FLOPPY_DISK(kObjHackingDisk,
        "floppy disk of hacking",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doHacking, 300, 20);
    ++n;
    /* enum: kObjAntivirus */
    NEW_FLOPPY_DISK(kObjAntivirus,
        "floppy disk of antivirus",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doAntiVirus, 300, 40);
    ++n;
    /* enum: kObjMatterCompiler */
    /* enum: kObjLastFloppyDisk = kObjMatterCompiler */
    NEW_FLOPPY_DISK(kObjMatterCompiler,
        "floppy disk of matter compiler",
        FloppyData[n].mDesc, FloppyData[n].mColor,
        &doMatterCompiler, 300, 20);
}
