#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <io.h>
#define getcwd _getcwd
#define access _access
#else
#include <unistd.h>
#endif
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#ifndef _WIN32
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif

#include "Global.h"
#include "Util.h"
#include "Map.h"
#include "AUI.h"
#include "NEUI.h"
#include "Monster.h"
#include "Hero.h"
#include "Game.h"

shInterface *I;
shMapLevel *Level;
int GodMode = 0;
char UserDir[PRIME_PATH_LENGTH];
shLogFile debug;

extern shFlags Flags;

static void
usage ()
{
    fprintf (stdout, "usage: prime [-u username]\n"
                     "       A science fiction themed roguelike game\n"
                     "       http://primetheroguelike.galeon.com\n");
}


int
main (int argc, char **argv)
{
    int rlimitresult = 0;
#ifndef _WIN32
	struct rlimit coredumpsize;

    coredumpsize.rlim_cur = RLIM_INFINITY;
    coredumpsize.rlim_max = RLIM_INFINITY;
    rlimitresult = setrlimit (RLIMIT_CORE, &coredumpsize);
#endif

    const char *name = NULL;
    char namebuf[HERO_NAME_LENGTH+2];
    int c;

    //FIXME: Here follows some ugly code interspersed with ifdefs.
    //TODO:  Create source file responsible for hiding portability issues.
    //FIXME: should probably stat() it and make sure it's a directory...
    if (access (SCOREDIR, O_RDWR)) {
        fprintf (stderr, "Couldn't open \"%s\" directory: %s.\n", SCOREDIR, strerror (errno));
        fprintf (stderr, "PRIME needs this directory for high scores and log.\n");
        exitPRIME (-1);
    }

    if (access (DATADIR, O_RDONLY)) {
        fprintf (stderr, "Couldn't open \"%s\" directory: %s.\n", DATADIR, strerror (errno));
        fprintf (stderr, "PRIME needs this directory for keybindings and help files.\n");
        exitPRIME (-1);
    }

#ifndef _WIN32
    if (USERDIR[0] == '~') { /* Insert user home directory. */
        const char *home = getenv ("HOME");
        if (!home) {
            fprintf (stderr, "HOME environment variable not present.\n");
            fprintf (stderr, "PRIME was configured to search for it and cannot run unless it is set.\n");
            exitPRIME (-1);
        }
        snprintf (UserDir, PRIME_PATH_LENGTH, "%s%s", home, &USERDIR[1]);
    } else {
        snprintf (UserDir, PRIME_PATH_LENGTH, "%s", USERDIR);
    }
#else
    snprintf (UserDir, PRIME_PATH_LENGTH, "%s", USERDIR);
#endif
#ifndef _WIN32
    if (access (UserDir, O_RDWR)) {
        mkdir (UserDir, S_IRUSR | S_IWUSR | S_IXUSR);
    }
#endif
    if (access (UserDir, O_RDWR)) {
        fprintf (stderr, "Couldn't open \"%s\" directory: %s.\n", UserDir, strerror (errno));
        fprintf (stderr, "PRIME needs this directory for saved games and user preferences.\n");
        exitPRIME (-1);
    }

    if (!debug.open ()) {
        exitPRIME (-1);
    }

    int g = 0, o = 0, d = 0;
    bool n = false; /* Use NotEye. */
#ifndef _WIN32
    while (-1 != (c = getopt (argc, argv, "u:N:dgno"))) {
        switch (c) {
        case 'd': d++; break;
        case 'g': g++; break;
        case 'o': o++; break;
        case 'n': n = true; break;
        case 'u':
            name = strdup (optarg);
            if (!nameOK (name)) {
                fprintf (stderr, "Sorry, can't play a game as \"%s\".\n",
                         name);
                return -1;
            }
            break;
        case 'N': break; /* Ignore. This option is for NotEye frontend. */
        case '?':
        default:
            usage ();
            return 0;
        }
    }
#else
    //windows doesn't have getopt()
    if (2 == argc && !strcmp("-god", argv[1]))
        d = g = o = 1;
#endif
/* FIXME: Ahem! On Linux running PRIME -dog will activate god mode. -- MB */
    if (d && g && o)
        GodMode = 1;

    utilInitialize ();

/*  if (n) {
        I = new shNotEyeInterface (argc, argv);
    } else {  */
        I = new shASCIIInterface ();
/*  }  */
    if (rlimitresult) {
        debug.log ("Couldn't adjust coredumpsize.");
    }

    initializeProfessions ();
    initializeMonsters ();

    I->showVersion ();

    if (!name) {
        bool invalid = false;

        name = getenv ("USER");
        if (!name)
            name = getenv ("USERNAME");

        do {
            I->p ("Welcome to PRIME!");
            if (invalid) I->p ("Invalid name!");
            if (!I->getStr (namebuf, HERO_NAME_LENGTH+1,
                           "What is your name?", name))
            {
                I->p ("Goodbye.");
                goto done;
            }
            I->pageLog ();
            invalid = true;
        } while (!nameOK (namebuf));
        name = namebuf;
    }

    I->loadOptions ();

    if (0 == loadGame (name) ||
        0 == newGame (name))
    {
        if (Flags.mKeySet[0] == '\0') I->keymapChoice ();
        I->p ("Welcome to \"PRIME\".  Press '?' or F1 for help.");
        I->p ("Please submit bug reports at psiweapon@gmail.com or michal@kast.net.pl");
        if (GodMode) {
            I->p ("God Mode is on.");
        }
        gameLoop ();
    }

    I->saveOptions ();
done:
    delete I;

    fprintf (stdout, "Thanks for playing!\n");
    fprintf (stdout, "Please submit bug reports at psiweapon@gmail.com or michal@kast.net.pl\n");
    exitPRIME (0);
}
